# PROBEER OOIT NOG EEN KEER OM DE countvec OP MAX TE ZETTEN OFZO
# https://github.com/paulgb/sklearn-pandas

# from sklearn.decomposition import PCA
import copy

import numpy as np
import pandas as pd
import scipy.sparse

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction import DictVectorizer
from sklearn.preprocessing import OneHotEncoder, robust_scale
from dateutil.parser import parse
from xtoy.utils import is_numeric
from xtoy.utils import is_integer

from xtoy.featurizers import SelectPolynomialByResidual
from xtoy.featurizers import RegexVectorizer


class DataFrameImputer(TransformerMixin):
    """
    Credits http://stackoverflow.com/a/25562948/1575066
    """

    def __init__(self):
        """Impute missing values.

        Columns of dtype object are imputed with the most frequent value
        in column.

        Columns of other types are imputed with mean of column.

        """
        self.fill = None

    @staticmethod
    def most_frequent(col):
        try:
            return col.value_counts().index[0]
        except IndexError:
            return 0

    def get_impute_val(self, col):
        if col.dtype == np.dtype("O"):
            val = self.most_frequent(col)
        elif col.dtype == np.dtype(float):
            val = col.mean()
        else:
            val = col.median()
        if isinstance(val, float) and np.isnan(val):
            val = 0
        return val

    def fit(self, X, y=None):
        self.fill = pd.Series([self.get_impute_val(X[c]) for c in X], index=X.columns)
        return self

    def transform(self, X, y=None):
        return X.fillna(self.fill, inplace=False)


class Featurizer(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        count_vectorizer_kwargs={"max_features": 15, "token_pattern": r"(?u)\b\w+\b"},
        max_unique_for_discrete=15,
        date_atts=("year", "month", "day", "weekday", "hour", "minute", "second"),
        regex_vectorizer=None,
        regex_patterns=None,
        sparse=True,
    ):
        self.count_vectorizer = CountVectorizer(**count_vectorizer_kwargs)
        self.regex_vectorizer = (
            regex_vectorizer or RegexVectorizer(regex_patterns) if regex_patterns else None
        )
        self.max_unique_for_discrete = max_unique_for_discrete
        self.one_hot_encoder = None
        # scary duplication
        self.feature_names_ = None
        self.feature_indices_ = None
        self.date_atts = date_atts
        self.sparse = True

    def fit(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        columns = list(X.columns)
        self.drop_vars = []
        self.count_vecs = []
        self.regex_vecs = []
        self.date_vars = []
        self.imputer = DataFrameImputer()
        self.ohe_indices = []
        self.numeric_indices = []
        self.missing_col_names = []
        self.feature_names_ = []
        self.feature_indices_ = []
        for col_name in X.columns[np.any(pd.isnull(X), axis=0)]:
            self.feature_names_.append(str(col_name) + "_missing")
            self.missing_col_names.append(col_name)
            self.feature_indices_.append(columns.index(col_name))
        X = self.imputer.fit_transform(X)
        self.ohe_indices = np.zeros(X.shape[1], dtype=bool)
        self.numeric_indices = np.zeros(X.shape[1], dtype=bool)
        self.drop_vars = np.zeros(X.shape[1], dtype=bool)
        self.date_vars = np.zeros(X.shape[1], dtype=bool)
        for i, col in enumerate(X):
            if len(set(X[col])) == 1:
                self.drop_vars[i] = True

        for i, col in enumerate(X):
            if self.drop_vars[i]:
                continue
            if is_numeric(X[col]):
                if len(set(X[col])) > 2:
                    self.feature_names_.append("{}_{}_continuous".format(col, i))
                else:
                    self.feature_names_.append("{}_{}_dummy".format(col, i))
                self.feature_indices_.append(i)
                num_unique = len(np.unique(X[col]))
                if is_integer(X[col]) and 3 <= num_unique <= self.max_unique_for_discrete:
                    self.ohe_indices[i] = True
                self.numeric_indices[i] = True
            # maybe date
            else:
                for x in X[col]:
                    try:
                        parse(x)
                    except ValueError:
                        break
                else:
                    self.date_vars[i] = True
                    date_var_names = [
                        "{}_date_{}".format(col, x) for j, x in enumerate(self.date_atts)
                    ]
                    self.feature_names_.extend(date_var_names)
                    self.feature_indices_.append(i)

        for i, (ohed, col) in enumerate(zip(self.ohe_indices, X)):
            if (
                self.count_vectorizer
                and not self.drop_vars[i]
                and not is_numeric(X[col])
                and not ohed
                and not self.date_vars[i]
            ):
                countvec = copy.copy(self.count_vectorizer)
                countvec.fit(["" if isinstance(x, float) else x for x in X[col]])
                self.count_vecs.append(countvec)
                cv_var_names = [
                    "{}_countvec_{}_{}_{}".format(col, x, i, j)
                    for j, x in enumerate(countvec.get_feature_names())
                ]
                self.feature_names_.extend(cv_var_names)
                self.feature_indices_.extend([i] * len(cv_var_names))
            else:
                self.count_vecs.append(False)

        for i, (ohed, col) in enumerate(zip(self.ohe_indices, X)):
            if (
                self.regex_vectorizer
                and not self.drop_vars[i]
                and not is_numeric(X[col])
                and not ohed
                and not self.date_vars[i]
            ):
                regex_vec = copy.copy(self.regex_vectorizer)
                regex_vec.dict_vectorizer = DictVectorizer(
                    **self.regex_vectorizer.dict_vectorizer.get_params()
                )
                regex_vec.fit(["" if isinstance(x, float) else x for x in X[col]])
                if regex_vec.get_feature_names():
                    self.regex_vecs.append(regex_vec)
                    rv_var_names = [
                        "{}_regex_vec_{}_{}_{}".format(col, x, i, j)
                        for j, x in enumerate(regex_vec.get_feature_names())
                    ]
                    self.feature_names_.extend(rv_var_names)
                    self.feature_indices_.extend([i] * len(rv_var_names))
                else:
                    self.regex_vecs.append(False)
            else:
                self.regex_vecs.append(False)

        self.one_hot_encoder = OneHotEncoder(handle_unknown="ignore")
        if np.any(self.ohe_indices):
            self.one_hot_encoder.fit(X[X.columns[self.ohe_indices]])
            # ridiculously complex OHE feature range names
            feat = set(self.one_hot_encoder.active_features_)
            for l, h, v in zip(
                self.one_hot_encoder.feature_indices_[:-1],
                self.one_hot_encoder.feature_indices_[1:],
                X.columns[self.ohe_indices],
            ):
                num_ohe_feat = len(feat.intersection(range(l, h)))
                self.feature_names_.extend(["{}_OHE_{}".format(v, j) for j in range(num_ohe_feat)])
                self.feature_indices_.extend([i] * num_ohe_feat)
        self.feature_names_ = np.array(self.feature_names_)
        self.feature_indices_ = np.array(self.feature_indices_)
        return self

    def transform(self, X, y=None):
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        res = []
        cvs = []
        rvs = []
        date_groups = []
        for col in self.missing_col_names:
            res.append(pd.isnull(X[col]))
        X = self.imputer.transform(X)
        for i, col in enumerate(X):
            if self.drop_vars[i] or self.count_vecs[i]:
                continue
            if self.numeric_indices[i]:
                res.append(X[col].astype(float))
            elif not self.ohe_indices[i] and self.date_vars[i]:
                prefix = str(col) + "__"
                dtimes = [parse(x) for x in X[col]]
                for a in self.date_atts:
                    dcol = [x.weekday() if a == "weekday" else getattr(x, a) for x in dtimes]
                    res.append(dcol)

        for cv, col in zip(self.count_vecs, X):
            if cv:
                cvs.append(
                    cv.transform(
                        ["" if isinstance(x, (np.int64, int, float)) else x for x in X[col]]
                    )
                )

        for rv, col in zip(self.regex_vecs, X):
            if rv:
                rvs.append(
                    rv.transform(
                        ["" if isinstance(x, (np.int64, int, float)) else x for x in X[col]]
                    )
                )

        combined = []
        if res:
            combined.append(scipy.sparse.coo_matrix(res).T)
        if cvs:
            combined.extend(cvs)

        if rvs:
            combined.extend(rvs)
        if np.any(self.ohe_indices):
            ohd = self.one_hot_encoder.transform(X[X.columns[self.ohe_indices]])
            combined = combined + [ohd]

        result = scipy.sparse.hstack(combined).tocsr()
        if not self.sparse:
            result = result.toarray()
        return result


Sparsify = Featurizer

# for k in train:
#     s = Featurizer()
#     XX = s.fit_transform(train[k])
#     assert len(s.feature_names_) == XX.shape[1]


# feature_important = pd.DataFrame(index=np.array(
#     s.feature_names_)[rfe.support_], data=clf.feature_importances_, columns=['importance'])
# feature_important = feature_important.sort_values(by=['importance'], ascending=True)
# feature_important.plot(kind='barh', stacked=True, color=[
#                        'cornflowerblue'], grid=False, figsize=(8, 5))


# import copy
# s = Featurizer()
# XX = s.fit_transform(train[::2])
# XX2 = s.transform(train[1::2])

# from sklearn.feature_selection import RFE

# rfe = RFE(LogisticRegression(), n_features_to_select=100, step=0.1)
# XX = rfe.fit_transform(XX, y[0::2])
# XX2 = rfe.transform(XX2)


# preds = []
# for clf in [RandomForestClassifier(1000, n_jobs=4),
#             MLPClassifier((100, 100)),
#             MLPClassifier((100, 100)),
#             KNeighborsClassifier(),
#             LogisticRegression()]:
#     clf.fit(XX, y[::2])
#     print(clf)
#     pred = clf.predict(XX2)
#     print(np.mean(pred == y[1::2]))
#     preds.append(pred)
#     clfs.append(clf)

# from scipy.stats import mode
# np.mean(mode(preds)[0].ravel() == y[1::2])



import warnings
import re
from collections import Counter

import numpy as np
from sklearn.base import TransformerMixin
from sklearn.feature_extraction import DictVectorizer


class RegexVectorizer(TransformerMixin):

    def __init__(self, regex_patterns=None, ignore_case=False, dict_vectorizer=DictVectorizer(),
                 binary=False):
        self.ignore_case = ignore_case
        self.binary = binary
        flags = ignore_case * re.IGNORECASE
        self.regex_patterns = regex_patterns
        self.regexes = {k: re.compile(v, flags=flags) for k, v in regex_patterns.items()}
        self.dict_vectorizer = dict_vectorizer

    def get_feature_names(self):
        return self.dict_vectorizer.get_feature_names()

    def fit(self, X, y=None):
        feats = self.featurize(X)
        self.dict_vectorizer.fit(feats)
        return self

    def transform(self, X, y=None):
        feats = self.featurize(X)
        return self.dict_vectorizer.transform(feats)

    @staticmethod
    def _merge_dicts(dict_args):
        result = Counter()
        for dictionary in dict_args:
            for k, v in dictionary.items():
                result[k] += v
        return result

    def featurize(self, X):
        if self.binary:
            feats = [{name: 1 for name, r in self.regexes.items() if r.search(x)} for x in X]
        else:
            feats = [self._merge_dicts([self._get_dict(name, r, x)
                                        for name, r in self.regexes.items()])
                     for x in X]
        return feats

    @staticmethod
    def _get_dict(name, r, text_value):
        matches = r.findall(text_value)
        if not matches:
            return Counter()
        return Counter(name + "_" + m for m in matches + ["_any"])

# X = ["I walked", "I am walking"]
# rv = RegexVectorizer({"walk": "walk[^ ]"})
# rv.fit_transform(X).todense()
# rv.featurize(X)


class SelectPolynomialByResidual():
    def __init__(self, clf, n_to_select="auto", test_size=0.4, feature_indices=None):
        self.clf = clf
        self.test_size = test_size
        self.feature_names = None
        self.feature_indices = None if feature_indices is None else feature_indices
        self.n_to_select = n_to_select
        self.best_coefs = None

    def transform(self, X, y=None):
        poly_features = []
        feature_indices = set(self.feature_indices)
        it = 0
        for i in range(X.shape[1]):
            for j in range(X.shape[1]):
                if i <= j:
                    if it in feature_indices:
                        poly_features.append(X[:, i] * X[:, j])
                    it += 1
        poly_features = np.vstack(poly_features).T
        return np.hstack((X, poly_features))

    def fit(self, X, y, var_names):
        # hack
        if self.feature_indices is not None:
            return self
        poly_features = []
        feature_names = []
        for i in range(X.shape[1]):
            for j in range(X.shape[1]):
                if i <= j:
                    feature_names.append(var_names[i] + "_times_" + var_names[j])
                    poly_features.append(X[:, i] * X[:, j])

        poly_features = np.vstack(poly_features).T
        feature_names = np.array(feature_names)

        if self.n_to_select == "auto":
            self.n_to_select = int(X.shape[1] / 10)

        self.clf.fit(X, y)
        y_residuals = (self.clf.predict(X) - y) ** 2

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            coefs = np.corrcoef(
                np.hstack((poly_features, y_residuals.reshape((-1, 1)))), rowvar=False)
        correlations = coefs[-1][:-1]

        # also consider weight on y
        coefs = np.corrcoef(np.hstack((poly_features, y_residuals.reshape(-1, 1))), rowvar=False)
        correlations = 2 * correlations + coefs[-1][:-1]

        indices = np.argsort(-np.abs(correlations))

        indices = indices[np.logical_not(np.isnan(indices))][:self.n_to_select]

        self.best_coefs = correlations[indices]
        self.feature_indices = indices
        self.feature_names = feature_names[indices]
        return self



# from sklearn.svm import SVC
import scipy
from sklearn import svm
from sklearn.decomposition import TruncatedSVD
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.linear_model import Ridge, RidgeClassifier
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor

from xtoy.utils import classification_or_regression


ridge_grid = {
    "estimator__alpha": [
        0.00001,
        0.001,
        0.01,
        0.05,
        0.1,
        0.3,
        0.5,
        0.8,
        1.0,
        1.4,
        1.7,
        10,
        100,
        1000,
        10000,
        100000,
    ]
}
# normalize seems bugged at prediction time!

# ridge_grid = {'estimator__alpha': [0.1, 1., 10.]}

INF = float("inf")
ridge_classification = {
    "clf": RidgeClassifier,
    "grid": ridge_grid,
    "name": "ridge",
    "max_complexity": INF,
}
ridge_regression = {"clf": Ridge, "grid": ridge_grid, "name": "ridge", "max_complexity": INF}

# ridge_regression = {'clf': svm.SVR, 'grid': {
#     'estimator__shrinking': [True, False],
#     'estimator__C': [0.001, 0.01, 0.1, 1, 10, 100, 1000],
#     'estimator__kernel': ['linear', 'poly', 'rbf', 'sigmoid']}}


rf_grid = {
    "estimator__max_features": ["sqrt", "auto", "log2", 0.5, 0.8, 0.9],
    "estimator__max_depth": [None, 5, 10, 20],
    "estimator__min_samples_leaf": [1, 5, 10],
    "estimator__min_samples_split": [2, 10, 20],
    "estimator__n_estimators": [1000],
}

knn_grid = {
    "estimator__n_neighbors": [1, 2, 3, 5, 10, 20],
    "estimator__leaf_size": [2, 3, 5, 10, 30, 50, 100],
    "estimator__p": [1, 2, 5, 10],
    "estimator__weights": ["uniform", "distance"],
}

rf_grid_classification = rf_grid.copy()
rf_grid_classification.update({"estimator__class_weight": ["balanced"]})

rf_classification = {
    "clf": RandomForestClassifier,
    "grid": rf_grid_classification,
    "name": "rf",
    "max_complexity": 100000 * 10,
}
rf_regression = {
    "clf": RandomForestRegressor,
    "grid": rf_grid,
    "name": "rf",
    "max_complexity": 100000 * 10,
}

knn_classification = {
    "clf": KNeighborsClassifier,
    "grid": knn_grid,
    "name": "knn",
    "max_complexity": 100000 * 100,
}
knn_regression = {
    "clf": KNeighborsRegressor,
    "grid": knn_grid,
    "name": "knn",
    "max_complexity": 100000 * 100,
}


options = {
    "regression": [ridge_regression, rf_regression, knn_regression],
    "classification": [ridge_classification, rf_classification, knn_classification],
}

xgb_grid = {
    "estimator__max_depth": [2, 3, 5, 10, 20],
    "estimator__min_child_weight": [1, 3, 5, 10, 15, 20],
    "estimator__learning_rate": [0.0001, 0.001, 0.001, 0.01, 0.1, 0.3, 0.5, 0.7, 1],
    "estimator__n_estimators": [100],
}


try:
    import xgboost

    xgb_classification = {
        "clf": xgboost.XGBClassifier,
        "grid": xgb_grid,
        "name": "xgb",
        "max_complexity": 100000 * 10,
    }
    xgb_regression = {
        "clf": xgboost.XGBRegressor,
        "grid": xgb_grid,
        "name": "xgb",
        "max_complexity": 100000 * 10,
    }
    options["classification"].append(xgb_classification)
    options["regression"].append(xgb_regression)
except ImportError:
    pass

lgb_grid = {
    "estimator__num_leaves": [2, 3, 5, 10, 20, 31, 50],
    "estimator__min_child_weight": [1, 3, 5, 10, 15, 20],
    "estimator__learning_rate": [0.0001, 0.001, 0.001, 0.01, 0.1, 0.3, 0.5, 0.7, 1],
}

try:
    import lightgbm as lgb

    lgb_classification = {
        "clf": lgb.LGBMClassifier,
        "grid": lgb_grid,
        "name": "lgb",
        "max_complexity": 100000 * 10,
    }
    lgb_regression = {
        "clf": lgb.LGBMRegressor,
        "grid": lgb_grid,
        "name": "lgb",
        "max_complexity": 100000 * 10,
    }
    options["classification"].append(lgb_classification)
    options["regression"].append(lgb_regression)
except ImportError:
    pass


def sparse_or_dense(X, RAM=None, magic=42):
    if scipy.sparse.issparse(X):
        return "sparse"
    else:
        # USING PAPER WE WILL GO TO DENSE ANYWAY
        return "dense"
    # size = np.prod(X.shape) if hasattr(X, 'shape') else len(X) * len(X[0])

    # # if N * num_feat * magic > RAM:
    # if size > 1000 ** 3:
    #     return 'sparse'
    # else:
    #     return 'dense'


def pick(X, y, cl_or_reg=None, opts=None):
    # i always choose first one now
    if opts is None:
        opts = options
    cl_or_reg = cl_or_reg if cl_or_reg else classification_or_regression(y)
    return opts[cl_or_reg]



import pandas as pd
import numpy as np
import scipy.stats

from sklearn.pipeline import Pipeline
from sklearn.preprocessing import Normalizer

from xtoy.evolutionary_search import EvolutionaryAlgorithmSearchCV as evo_search
from xtoy.prep import Featurizer
from xtoy.classifiers import pick
from xtoy.classifiers import classification_or_regression

from xtoy.scorers import f1_weighted_scorer
from xtoy.scorers import mse_scorer

from xtoy.multi_output import MOR, MOC
from xtoy.utils import get_cv_splits
from xtoy.utils import calculate_complexity

from sklearn.neighbors.base import NeighborsBase

try:
    import pickle
except (ValueError, SystemError, ImportError):
    pass


class Toy:
    def __init__(
        self, cv=get_cv_splits, scoring=None, n_jobs=1, cl_or_reg=None, sparse=True, **kwargs
    ):
        self.cv = get_cv_splits
        self.evo = None
        self.scoring = scoring
        self.n_jobs = n_jobs
        self.kwargs = kwargs
        self.cl_or_reg = cl_or_reg
        self.featurizer = Featurizer(sparse=sparse)
        self._feature_name = None
        self.evos = []

    def get_models(self, X, y):
        return pick(X, y, self.cl_or_reg)

    def get_pipeline(self, clf):
        return Pipeline(
            steps=[
                # ('tsvd', TruncatedSVD()),  # this one also has to have % top features chosen
                # ('feature_selection', SelectFromModel(Ridge())),
                ("scaler", Normalizer()),
                ("estimator", clf()),
            ]
        )

    def handle_multi_output(self, y, name, clf):
        y = np.array(y)
        if len(y.shape) > 1 and y.shape[1] > 1:
            if name == "xgb":
                return None
            MO = MOC if "Classif" in clf.__name__ else MOR
            return lambda: MO(clf())
        return clf

    def fit(self, X, y):
        evos = []
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        self._feature_name = np.array(list(X.columns))
        if isinstance(y, pd.DataFrame):
            y = np.array(y)
        elif hasattr(y, "__array__"):
            y = y.__array__()
        elif len(y.shape) > 1 and y.shape[1] == 1:
            y = y.ravel()
        X = pd.DataFrame(self.featurizer.fit_transform(X).A)
        self.cl_or_reg = self.cl_or_reg or classification_or_regression(y)
        cl_or_reg = self.cl_or_reg
        if self.scoring is None:
            self.scoring = [f1_weighted_scorer, mse_scorer][cl_or_reg != "classification"]
        print(self.scoring)

        complexity = calculate_complexity(X, y, cl_or_reg)
        for model in self.get_models(X, y):
            try:
                print("estimator:", model["name"])
                grid = model["grid"]
                if complexity > model["max_complexity"]:
                    continue
                clf = self.handle_multi_output(y, model["name"], model["clf"])
                if clf is None:
                    continue
                pipeline = self.get_pipeline(clf)
                unique_combinations = np.prod(list(map(len, grid.values())))
                print("unique_combinations", unique_combinations)
                kwargs = self.kwargs.copy()
                if "population_size" not in self.kwargs:
                    kwargs["population_size"] = np.clip(int(unique_combinations / 100), 5, 10)
                if "generations_number" not in kwargs:
                    kwargs["generations_number"] = np.clip(int(unique_combinations / 10), 10, 50)

                evo = evo_search(
                    pipeline, grid, scoring=self.scoring, cv=self.cv, n_jobs=self.n_jobs, **kwargs
                )
                evo.fit(X, y)
                evos.append((evo.best_score_, evo))
            except KeyboardInterrupt:
                if not evos:
                    print("Stopped by user. No models finished trained; failed to fit.")
                    raise
                print("Stopped by user. {} models trained.".format(len(evos)))
        self.evos = evos
        self.best_evo = sorted(self.evos, key=lambda x: x[0])[-1][1]
        # import warnings
        # warnings.warn("best: {}".format(self.best_evo.best_estimator_))
        print("best: {}".format(self.best_evo.best_estimator_))
        return self.best_evo.best_estimator_

    def predict(self, X):
        X = self.featurizer.transform(X).A
        return self.best_evo.predict(X)

    def predict_proba(self, X):
        X = self.featurizer.transform(X).A
        return self.best_evo.predict_proba(X)

    def ensemble_predict(self, X):
        fn = scipy.stats.mode if self.cl_or_reg == "classification" else np.mean
        return fn([x[1].predict(X) for x in self.evos], axis=0)[0][0]

    def ensemble_importances_(self, X, y):
        for score, clf in self.evos:
            clf.estimator.fit(X, y)
        Z = [x[0] * self.feature_importances_(x[1].estimator.steps[-1][1]) for x in self.evos]
        weight_sum = sum([x[0] for x in self.evos])
        return np.sum(Z, axis=0) / weight_sum

    def score(self, X, y):
        X = self.featurizer.transform(X).A
        return self.best_evo.best_estimator_.score(X, y)

    # def baselines():
    #     f1_weighted_score

    def best_model_pickle(self):
        return pickle.dumps(self.best_pipeline_)

    def feature_importances_(self, clf):
        if hasattr(clf, "estimator"):
            clf = clf.estimator
        if hasattr(clf, "feature_importances_"):
            weights = clf.feature_importances_
        elif hasattr(clf, "coef_"):
            weights = np.abs(clf.coef_)
        elif isinstance(clf, NeighborsBase):
            weights = np.ones(len(self.featurizer.feature_names_))
        else:
            raise ValueError("No importances could be computed (requires a different classifier).")
        weights = np.abs(weights)
        if len(weights.shape) > 1:
            weights = weights.mean(axis=0)
        weights = weights / np.sum(weights)
        assert np.isclose(weights.sum(), 1)
        return weights

    @property
    def feature_names_(self):
        return self.featurizer.feature_names_

    @property
    def feature_indices_(self):
        return self.featurizer.feature_indices_

    def best_features_(self, importances=None, n=10, aggregation=np.max):
        # a bit annoying that aggregation makes different shape if aggregation=None
        # this is whether interested in original features, or post processing.
        # maybe split this in 2 functions
        """
        Default is to use the feature importances from the best model.
        If importances is not None, it is expected to be an array with weights for the features.

        By default it will aggregate the importances for e.g. text and categorical features.
        If aggregation is set to None, it will instead print the raw importances of converted X

        Feature weights sum to 1. """
        if importances is None:
            importances = self.feature_importances_(self.best_evo.best_estimator_.steps[-1][1])
        if aggregation is None:
            data = list(zip(importances, self.feature_names_))
        else:
            pdata = pd.DataFrame(
                {"features": self._feature_name[self.feature_indices_], "importances": importances}
            )
            agg = pdata.groupby(["features"]).agg(aggregation)
            data = list(zip(agg["importances"].values, agg.index))
        return sorted(data)[-n:]

    @property
    def best_pipeline_(self):
        return self.best_evo.best_estimator_




from sklearn.multioutput import MultiOutputRegressor, MultiOutputClassifier


class MOR(MultiOutputRegressor):
    def get_params(self, deep=False):
        params = super(MOR, self).get_params(deep=deep)
        params = {k.replace("estimator__", ""): v for k, v in params.items()}
        return params


class MOC(MultiOutputClassifier):
    def get_params(self, deep=False):
        params = super(MOR, self).get_params(deep=deep)
        params = {k.replace("estimator__", ""): v for k, v in params.items()}
        return params



from setuptools import setup
from setuptools import find_packages

MAJOR_VERSION = '0'
MINOR_VERSION = '5'
MICRO_VERSION = '57'
VERSION = "{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION)

setup(
    name="xtoy",
    version=VERSION,
    description="get xtoyed predictions from raw data",
    url="https://github.com/kootenpv/xtoy",
    author="Pascal van Kooten",
    author_email="kootenpv@gmail.com",
    license="MIT",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "scipy",
        "pandas",
        "scikit-learn>=0.18.1,<0.20",
        "xgboost==0.72.1",
        "lightgbm==2.2.2",
        "deap",
        "bitstring",
    ],
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Microsoft",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ],
    keywords=["data science", "machine learning", "genetic programming"],
    zip_safe=False,
    platforms="any",
)
